/*
   Ray Tracing Triangles Demo.
   Created by the Programming Ace.
   www.UltimateGameProgramming.com

   The purpose of this demo is to build off of the first two ray tracing
   demo by adding the ability to ray trace triangles.  This demo will
   create two triangles that will act as a floor.  To do this we will
   need a new class, the Triangle class, and we will need it give it
   its own intersect function and normal function.  I've changed the
   code a bit to make use of classes for the shapes instead of structs.
   I also added a shape class to hold a shape (like triangle or sphere).
   Of course there are better ways to do this but I was aiming for simplicity
   so that learning what is going on is easier to see.  Also I didn't want
   the code changing too much from the original two.
*/


// Demo: Ray Tracing Triangles.


// All the necessary includes...
#include<iostream>
#include"Shape.h"
#include"RayTracer.h"
#include"Light.h"
#include"TgaImage.h"

using namespace std;

// Width/Height resolution and total number of objects.
#define WIDTH_RES 640
#define HEIGHT_RES 480
#define TOTAL_SHAPES 4


// Function prototype...
bool InitializeRT();
void ClearBuffer(unsigned char *buffer, int size, int r, int g, int b);
void Shutdown();


// Primary render array and the look up array of directions.
unsigned char *primaryBuffer = 0;
CVector4 *directionList = 0;

// List of objects and light source.
Shape shapes[TOTAL_SHAPES];
CLight light;


int main()
{
   cout << "Ultimate Game Programming Ray Tracer..." << endl;
   cout << "Created by Allen Sherrod..." << endl << endl;

   // Initialize all arrays and data nescessary.  If it fails then we can't
   // perform our ray tracing.
   if(!InitializeRT())
      {
         cout << "Initialization failed!  Exiting." << endl;
         return 0;
      }

   cout << "Clearing primary buffer..." << endl;

   // Clear the rendered scene to a default color (I choose black).
   ClearBuffer(primaryBuffer, WIDTH_RES * HEIGHT_RES * 3, 0, 0, 0);
   
   cout << "Primary buffer cleared and ready to be drawn to..." << endl << endl;
   cout << "Starting trace." << endl;

   // Trace through the scene and create the rendered scene.
	TraceScene(shapes, TOTAL_SHAPES, light, primaryBuffer, directionList,
              WIDTH_RES, HEIGHT_RES);

   cout << "Trace over." << endl << endl;
   
   cout << "Preparing to save scene as RayTracedScene.tga..." << endl;
   
   // Take our ray traced scene and save it as a .tga image file.
   WriteTGA("RayTracedScene.tga", WIDTH_RES, HEIGHT_RES, primaryBuffer);

   cout << "Traced scene saved...shut down starting..." << endl << endl;

   // Release all resources and data we used.
   Shutdown();

   cout << "Exiting..." << endl;

   return 1;
}


bool InitializeRT()
{
   // Release all previous data.
   Shutdown();

   // Allocate rendering buffer.
	primaryBuffer = new unsigned char[WIDTH_RES * HEIGHT_RES * 3];

   // Error checking...
   if(primaryBuffer == NULL)
      return false;

	// Allocate the ray directions.
	directionList = GenerateRayDirections(WIDTH_RES, HEIGHT_RES, 255);

   // Error checking...
   if(directionList == NULL)
      return false;

   // Setup the light.
   light.position = CVector4(75, 150, -350);
   light.r = 255; light.g = 255; light.b = 255;

   // Initialize the shapes we are using.
   shapes[0].type = SPHERE;
   shapes[0].sphere = new Sphere(-120, -50, 0, 100, 0, 0, 255);

   shapes[1].type = SPHERE;
   shapes[1].sphere = new Sphere(120, -50, 0, 100, 0, 255, 0);
   
   shapes[2].type = TRIANGLE;
   shapes[2].triangle = new Triangle(CVector4(-800, -150, -200),
                                     CVector4(-800, -150, 200),
                                     CVector4(800, -150, 200), 255, 255, 255);
   
   shapes[3].type = TRIANGLE;
   shapes[3].triangle = new Triangle(CVector4(800, -150, 200),
                                     CVector4(800, -150, -200),
                                     CVector4(-800, -150, -200), 255, 255, 255);

   return true;
}


void ClearBuffer(unsigned char *buffer, int size, int r, int g, int b)
{
   // Loop through the entire array and set the buffer to the specified color.
   for(int p = 0; p < size; p += 3)
      {
         buffer[p + 0] = r; buffer[p + 1] = g; buffer[p + 2] = b;
      }
}


void Shutdown()
{
   // Release all resources we used.
   if(directionList != NULL) delete[] directionList;
   if(primaryBuffer != NULL) delete[] primaryBuffer;


   for(int i = 0; i < TOTAL_SHAPES; i++)
      {
         if(shapes[i].type == SPHERE && shapes[i].sphere != NULL)
            delete shapes[i].sphere;
         else if(shapes[i].type == TRIANGLE && shapes[i].triangle != NULL)
            delete shapes[i].triangle;
      }
}


// Recap:

// With something more than just two sphere the scene looks a lot better.
// We still have a long way to go before this ray tracer can create some
// very impressive scenes but we are on the way.  Next up, planes.


// Copyright February 2005
// All Rights Reserved!
// Allen Sherrod
// ProgrammingAce@UltimateGameProgramming.com
// www.UltimateGameProgramming.com